bug-coreutils
[Top][All Lists]
Advanced

[Date Prev][Date Next][Thread Prev][Thread Next][Date Index][Thread Index]

fix for long-standing bug in cp --sparse


From: Jim Meyering
Subject: fix for long-standing bug in cp --sparse
Date: Sat, 19 Aug 2006 17:19:05 +0200

Alan, thanks for the report and patch.
FYI, I expect all !HAVE_FTRUNCATE code to be removed sometime in 2007.

2006-08-19  Jim Meyering  <address@hidden>

        * NEWS: Fix cp --sparse so that it preserves tail-end sparseness, even
        when the file's apparent size is not a multiple of its block size.
        * src/copy.c (copy_reg): Don't write a NUL before calling ftruncate.
        For some file sizes, writing that single byte would unnecessarily
        waste a few file blocks.  That write may have been necessary in the
        early days of Linux, but now, removing it should be safe.
        Based on a patch by Alan Curry: <http://bugs.debian.org/370792>
        * tests/cp/sparse: New test for the above.
        * tests/cp/Makefile.am (TESTS): Add sparse.

        * tests/sparse-file: New file, essence factored out of...
        * tests/du/8gb: ... here.  Use the new script.

Index: NEWS
===================================================================
RCS file: /fetish/cu/NEWS,v
retrieving revision 1.402
diff -u -r1.402 NEWS
--- NEWS        17 Aug 2006 19:58:18 -0000      1.402
+++ NEWS        19 Aug 2006 10:28:18 -0000
@@ -4,6 +4,10 @@
 
 ** Bug fixes
 
+  cp --sparse preserves sparseness at the end of a file, even when
+  the file's apparent size is not a multiple of its block size.
+  [introduced with the original design, in fileutils-4.0r, 2000-04-29]
+
   df (with a command line argument) once again prints its header
   [introduced in coreutils-6.0]
 
Index: tests/cp/sparse
===================================================================
RCS file: tests/cp/sparse
diff -N tests/cp/sparse
--- /dev/null   1 Jan 1970 00:00:00 -0000
+++ tests/cp/sparse     19 Aug 2006 06:12:16 -0000
@@ -0,0 +1,57 @@
+#!/bin/sh
+# Test cp --sparse=always
+
+# Copyright (C) 2006 Free Software Foundation, Inc.
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+# 02110-1301, USA.
+
+if test "$VERBOSE" = yes; then
+  set -x
+  cp --version
+fi
+
+. $srcdir/../envvar-check
+. $srcdir/../sparse-file
+
+pwd=`pwd`
+t0=`echo "$0"|sed 's,.*/,,'`.tmp; tmp=$t0/$$
+trap 'status=$?; cd $pwd; chmod -R u+rwx $t0; rm -rf $t0 && exit $status' 0
+trap '(exit $?); exit $?' 1 2 13 15
+
+framework_failure=0
+mkdir -p $tmp || framework_failure=1
+cd $tmp || framework_failure=1
+
+# Create a sparse file.
+# It has to be at least 128K in order to be sparse on some systems.
+# Make its size one larger than 128K, in order to tickle the
+# bug in coreutils-6.0.
+size=`expr 128 \* 1024 + 1`
+dd bs=1 seek=$size of=sparse < /dev/null 2> /dev/null || framework_failure=1
+
+if test $framework_failure = 1; then
+  echo "$0: failure in testing framework" 1>&2
+  (exit 1); exit 1
+fi
+
+fail=0
+
+cp --sparse=always sparse copy || fail=1
+
+# Ensure that the copy has the same block count as the original.
+test `stat --printf %b sparse` = `stat --printf %b copy` || fail=1
+
+(exit $fail); exit $fail
Index: tests/cp/Makefile.am
===================================================================
RCS file: /fetish/cu/tests/cp/Makefile.am,v
retrieving revision 1.34
diff -u -r1.34 Makefile.am
--- tests/cp/Makefile.am        17 Aug 2006 19:58:25 -0000      1.34
+++ tests/cp/Makefile.am        19 Aug 2006 05:52:48 -0000
@@ -21,6 +21,7 @@
 AUTOMAKE_OPTIONS = 1.1 gnits
 
 TESTS = \
+  sparse \
   link-no-deref \
   cp-deref \
   acl \
Index: tests/sparse-file
===================================================================
RCS file: tests/sparse-file
diff -N tests/sparse-file
--- /dev/null   1 Jan 1970 00:00:00 -0000
+++ tests/sparse-file   19 Aug 2006 10:54:49 -0000
@@ -0,0 +1,33 @@
+# -*- sh -*-
+# Does the current (working-dir.) file system support sparse files?
+
+# Copyright (C) 2006 Free Software Foundation, Inc.
+
+# This program is free software; you can redistribute it and/or modify
+# it under the terms of the GNU General Public License as published by
+# the Free Software Foundation; either version 2 of the License, or
+# (at your option) any later version.
+
+# This program is distributed in the hope that it will be useful,
+# but WITHOUT ANY WARRANTY; without even the implied warranty of
+# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
+# GNU General Public License for more details.
+
+# You should have received a copy of the GNU General Public License
+# along with this program; if not, write to the Free Software
+# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
+# 02110-1301, USA.
+
+# Test whether we can create a sparse file.
+# For example, on Darwin6.5 with a file system of type hfs, it's not possible.
+# NTFS requires 128K before a hole appears in a sparse file.
+t=sparse.$$
+dd bs=1 seek=128K of=$t < /dev/null 2> /dev/null
+set x `du -sk $t`
+kb_size=$2
+rm -f $t
+if test $kb_size -ge 128; then
+  echo "$0: skipping this test, since this file system doesn't support" \
+    sparse files 1>&2
+  (exit 77); exit 77
+fi
Index: tests/du/8gb
===================================================================
RCS file: /fetish/cu/tests/du/8gb,v
retrieving revision 1.10
diff -u -r1.10 8gb
--- tests/du/8gb        17 Aug 2006 19:58:26 -0000      1.10
+++ tests/du/8gb        19 Aug 2006 05:51:17 -0000
@@ -2,7 +2,7 @@
 # Ensure that du does not rely on narrow types like size_t for
 # file sizes or sums.
 
-# Copyright (C) 2003, 2005 Free Software Foundation, Inc.
+# Copyright (C) 2003, 2005, 2006 Free Software Foundation, Inc.
 
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -25,6 +25,7 @@
 fi
 
 . $srcdir/../envvar-check
+. $srcdir/../sparse-file
 
 pwd=`pwd`
 t0=`echo "$0"|sed 's,.*/,,'`.tmp; tmp=$t0/$$
@@ -40,19 +41,6 @@
   (exit 1); exit 1
 fi
 
-# If this file system doesn't support sparse files,
-# don't try to create a file that'd end up consuming 8GB.
-# This happens on Darwin6.5 with a file system of type `hfs'.
-# NTFS requires 128K before a hole appears in a sparse file.
-dd bs=1 seek=128K of=t < /dev/null 2> /dev/null
-set x `du -sk t`
-if test "$2" -ge 128; then
-  echo "$0: skipping this test, since this file system doesn't support" 1>&2
-  echo "$0: sparse files and this test requires a file with an apparent" 1>&2
-  echo "$0: size of 8GB" 1>&2
-  (exit 77); exit 77
-fi
-
 dd bs=1 seek=8G of=big < /dev/null 2> /dev/null
 if test $? != 0; then
   echo "$0: cannot create a file large enough for this test; possibly" 1>&2




reply via email to

[Prev in Thread] Current Thread [Next in Thread]
Index: src/copy.c
===================================================================
RCS file: /fetish/cu/src/copy.c,v
retrieving revision 1.204
diff -u -r1.204 copy.c
--- src/copy.c  17 Aug 2006 15:35:00 -0000      1.204
+++ src/copy.c  19 Aug 2006 05:35:24 -0000
@@ -461,21 +461,21 @@
            }
        }
 
-      /* If the file ends with a `hole', something needs to be written at
-        the end.  Otherwise the kernel would truncate the file at the end
-        of the last write operation.  */
+      /* If the file ends with a `hole', we need to do something to record
+        the length of the file.  On modern systems, calling ftruncate does
+        the job.  On systems without native ftruncate support, we have to
+        write a byte at the ending position.  Otherwise the kernel would
+        truncate the file at the end of the last write operation.  */
 
       if (last_write_made_hole)
        {
-#if HAVE_FTRUNCATE
-         /* Write a null character and truncate it again.  */
-         if (full_write (dest_desc, "", 1) != 1
-             || ftruncate (dest_desc, n_read_total) < 0)
-#else
-         /* Seek backwards one character and write a null.  */
-         if (lseek (dest_desc, (off_t) -1, SEEK_CUR) < 0L
-             || full_write (dest_desc, "", 1) != 1)
-#endif
+         if (HAVE_FTRUNCATE
+             ? /* ftruncate sets the file size,
+                  so there is no need for a write.  */
+             ftruncate (dest_desc, n_read_total) < 0
+             : /* Seek backwards one character and write a null.  */
+             (lseek (dest_desc, (off_t) -1, SEEK_CUR) < 0L
+              || full_write (dest_desc, "", 1) != 1))
            {
              error (0, errno, _("writing %s"), quote (dst_name));
              return_val = false;